/* gui.c */

/* VIS-5D version 3.0 */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <stdio.h>
#include <math.h>
#include "../lui2/lui.h"
#include "cursor.h"
#include "cw.h"
#include "globals.h"
#include "graphics.h"
#include "labels.h"
#include "matrix.h"
#include "misc.h"
#include "queue.h"
#include "render.h"
#include "save.h"
#include "sb.h"
#include "slice.h"
#include "topo.h"


#define ABS(x)  ( (x) < 0 ? -(x) : (x) )
#define MM_NORMAL 1
#define MM_TRAJ   2
#define MM_LABEL  3
#define MM_SLICE  4


static Display  *wdpy, *gdpy; /* widget & graphics displays */
/*static Display  *dpy;	*/      /* the application's display */
static Window  mainwin;	      /* the main 3-D viewing window */
static Window  CpWindow;      /* the control panel window */
static int Perspec = 0;
static int MouseMode;




/*** make_gui *********************************************************
   Setup all the Graphical User Interface stuff.
**********************************************************************/
int make_gui( wdpy_name, fontname, fontsize, fullflag )
char *wdpy_name;
char fontname[];
int fontsize;
int fullflag;
{
   int width, height, xpos, ypos;

   /* open the graphics display on default */
   gdpy = XOpenDisplay( NULL );
   if (!gdpy) {
      printf("Unable to open default display\n");
      exit(1);
   }

   /* open widget display */
   if (wdpy_name==NULL) {
      wdpy = gdpy;
   }
   else {
      wdpy = XOpenDisplay( wdpy_name );
      if (!wdpy) {
	 printf("Unable to open display %s\n", wdpy_name );
	 exit(1);
      }
   }

   /* make the widgets */
   create_widgets();


   /* make 3-D window */
   if (fullflag) {
      width = ScrWidth;
      height = ScrHeight;
      xpos = ypos = 0;
   }
   else {
      height = width = ScrWidth - 420;
      xpos = 410;
      ypos = 10;
   }
   mainwin = make_window( gdpy, ProgramName, xpos,ypos, width, height,
			 fontname, fontsize );


   XRaiseWindow( wdpy, CpWindow );

#ifdef SPACEBALL
   initialize_spaceball();
#endif

}



sync()
{
   XSync( wdpy, 0 );
}


#define BW 6
#define BUTTONSIZE 90   /* Width of regular buttons */
#define RADIOSIZE 120   /* Width of radio buttons */
#define CP_WIDTH  384
#define NBCOL 4

/* depends on window border size: */
#ifdef stellar
#  define SPACE 15
#else
#  define SPACE 40
#endif



static char *Copyright[] = {
   "             VIS-5D version 3.0               ",
   "        Copyright (C) 1990, 1991, 1992        ",
   "  Bill Hibbard, Brian Paul & Andre Battaiola  "
};

static char *ModeInfo1[] = {
    " Change the Viewing Angle ",
    "                          ",
    "      Mouse Buttons       ",
    "--------------------------",
    " rotate | zoom & | trans- ",
    "  view  |  clip  |  late  "
};

static char *ModeInfo2[] = {
    " Make & View Trajectories ",
    "                          ",
    "       Mouse Buttons      ",
    "--------------------------",
    " rotate |  make  |  move  ",
    "  view  | trajec | cursor "
};

static char *ModeInfo3[] = { 
    " Make & Edit Text Labels  ",
    "                          ",
    "       Mouse Buttons      ",
    "--------------------------",
    "  make  |  move  | delete ",
    "  label |  label | label  "
};

static char *ModeInfo4[] = {
    "  Change Slice Positions  ",
    "                          ",
    "       Mouse Buttons      ",
    "--------------------------",
    " rotate | zoom & | move   ",
    "  view  |  clip  | slice  "
};


static char *Columns[] = {
   "  Iso-     Contour Slice      Colored Slice  ",
   "  Surf    Horiz.    Vert.    Horiz.    Vert.  "
};


static char *TitleTraj[] = {
    "   Interactive Wind Trajectories    ",
    " LFT-Display, CNT-Select, RT-Color  "
};


static char *CloneTitle[] = {"Select variable to clone:"};


static Window  ColorWindow, IsoWindow,
  HSliceWindow, VSliceWindow, RGBWindow, TrajWindow,
  VerifyWindow, AlertWindow, CloneWindow, WindWindow;
static int CpHeight, cpx = 10, cpy = 10;
static int IsosurfHeight, IsosurfWidth;
static int HSliceWidth, HSliceHeight;
static int VSliceWidth, VSliceHeight;
static int RGBWidth=CP_WIDTH, RGBHeight=200;
static int TrajWidth, TrajHeight;
static int VerifyHeight, AlertHeight;
static int CloneWidth, CloneHeight;
static int WindWidth, WindHeight;
static LUI_SLIDER *IsoSlider;
static LUI_SLIDER *red_slider, *green_slider, *blue_slider, *trans_slider;
static int ColorHeight, ColorWidth;
static unsigned long  *ColorPtr ;
static LUI_BUTTON *Last_button;
static LUI_BUTTON *pretty_button;
static LUI_BUTTON *Button[MAXPARMS*5+TRAJGROUPS+2*WINDSLICES];
static int NumButtons;

static char *HSliceIntStr[2];
static char *HSliceLabStr[2];
static char *VSliceIntStr[2];
static char *VSliceLabStr[2];
static char *WindLabStr[2];
static char *WindLab2Str[2];
static char *WindDensityStr[2];
static char *WindScaleStr[2];
static char TrajStepLabel[50] = "Step:  ";
static char *TrajStepValue[2];
static char TrajLengthLabel[50] = "Length: ";
static char *TrajLengthValue[2];
static char *VerifyMessage[2];
static char *AlertMessage[3];


/* 'current' isosurf, hor, ver slice, traj, wind slices */
static int ip3d = -1, ip2d = -1, ipxs = -1, iphs = -1, ipvs = -1,
  iptraj = 0, iphw = -1, ipvw = -1, iptopo;


int main_process(), button_cb(),
  red_cb(), blue_cb(), green_cb(), trans_cb(),
  isosurface_cb(), hslice_cb(), vslice_cb(),
  anim_cb(), step_cb(), proj_cb(), exit_cb(), mode_cb(), topo_cb(),
  map_cb(), view_cb(), contnum_cb(), coord_cb(), hwind_cb(), vwind_cb(),
  save_cb(), restore_cb(), box_cb(), clock_cb(),
  traj_cb(), ribbon_cb(), dellast_cb(), delset_cb(), trajstep_cb(),
  trajlength_cb(), ok_cb(), cancel_cb(), debug_cb(), pretty_cb(),
  clone_cb(), do_clone_cb(), newsurf_cb(), windscale_cb(), winddensity_cb(),
  reverse_cb(), info_cb();

extern void terminate();




static create_widgets()
{
   int x, y, w, h, i, j, index, bottom;
   char label[20], rows[1000];
   Window parent;
   LUI_BUTTON *lb;

   /* allocate memory for type-in labels */
   HSliceLabStr[0] = (char *) calloc( 100, sizeof(char) );
   HSliceIntStr[0] = (char *) calloc( 100, sizeof(char) );
   HSliceLabStr[1] = HSliceIntStr[1] = NULL;

   VSliceLabStr[0] = (char *) calloc( 100, sizeof(char) );
   VSliceIntStr[0] = (char *) calloc( 100, sizeof(char) );
   VSliceLabStr[1] = VSliceIntStr[1] = NULL;

   WindLabStr[0] = (char *) calloc( 100, sizeof(char) );
   WindLabStr[1] = NULL;
   WindLab2Str[0] = (char *) calloc( 100, sizeof(char) );
   strcpy( WindLab2Str[0], "Density:" );
   WindLab2Str[1] = NULL;
   WindDensityStr[0] = (char *) calloc( 100, sizeof(char) );
   WindDensityStr[1] = NULL;
   WindScaleStr[0] = (char *) calloc( 100, sizeof(char) );
   WindScaleStr[1] = NULL;

   /* initialize type-ins */
   TrajStepValue[0] = (char *) calloc( 100, sizeof(char) );
   sprintf( TrajStepValue[0], "%6.2f", TrajStep );
   TrajLengthValue[0] = (char *) calloc( 100, sizeof(char) );
   sprintf( TrajLengthValue[0], "%6.2f", TrajLength );
   TrajStepValue[1] = TrajLengthValue[1] = NULL;

   /* verification and alert window message */
   VerifyMessage[0] = (char *) calloc( 100, sizeof(char) );
   AlertMessage[0] = (char *) calloc( 100, sizeof(char) );
   AlertMessage[1] = (char *) calloc( 100, sizeof(char) );
   VerifyMessage[1] = AlertMessage[2] = NULL;


   /* set the Widget Hue before initializing */
   LUI_Color( 0xffffff );
   LUI_Initialize( ProgramName, wdpy );

   /* Create control panel window */
   CpWindow = LUI_CreateWindowAt( LUI_RootWindow, /* parent */
                  cpx, cpy, /* position */ 100, 100 /* temporary size */
               );
   CpHeight = BW;


   /******************************************************************/
   /* Create the copyright label.                                    */
   /******************************************************************/
   LUI_LabelOpen ( "Copyright",      /* name */
             3, Copyright,           /* text strings */
             CpWindow,               /* parent window */
             BW, BW,                 /* origin relative to parent */
             CP_WIDTH, 50,           /* dimensions */
             BW,                     /* border size */
             NULL                    /* callback function */
             );
   LUI_LabelClose("Copyright");
   LUI_LabelVisible("Copyright", 1);
   LUI_LabelHighlight("Copyright", 0);
   LUI_LabelQuery("Copyright", &x, &y, &w, &h);
   CpHeight += h;


   /************************************************************/
   /* Create the main button pad.                              */
   /************************************************************/
   LUI_ButtonPadOpen ("Buttons", /* name */
		 CpWindow,       /* parent window */
		 BW, CpHeight,   /* origin relative to parent */
		 BW,             /* border size */
		 0x000000        /* background color */
		 );

   /** Row 1 **/
   if (NumTimes>1) {
      LUI_ButtonCreate("ANIMATE",LUI_BUTTON_TOGGLE, 0,0,BUTTONSIZE,0, anim_cb);
      LUI_ButtonCreate("STEP",   LUI_BUTTON_ONESHOT,1,0,BUTTONSIZE,0, step_cb);
   }
   else {
      LUI_ButtonCreate("", LUI_BUTTON_ONESHOT, 0,0,BUTTONSIZE,0, NULL);
      LUI_ButtonCreate("", LUI_BUTTON_ONESHOT, 1,0,BUTTONSIZE,0, NULL);
   }
   LUI_ButtonCreate("CLONE",LUI_BUTTON_ONESHOT,2,0,BUTTONSIZE,0, clone_cb);
   LUI_ButtonCreate("EXIT", LUI_BUTTON_ONESHOT,3,0,BUTTONSIZE,0, exit_cb);

   /** Row 2 **/
   LUI_ButtonCreate("NEWSURF",LUI_BUTTON_ONESHOT,0,1,BUTTONSIZE,0, newsurf_cb);
   LUI_ButtonCreate("TOP",    LUI_BUTTON_ONESHOT,1,1,BUTTONSIZE,1, view_cb);
   LUI_ButtonCreate("SOUTH",  LUI_BUTTON_ONESHOT,2,1,BUTTONSIZE,2, view_cb);
   LUI_ButtonCreate("WEST",   LUI_BUTTON_ONESHOT,3,1,BUTTONSIZE,3, view_cb);

   /** Row 3 **/
   if (TopoFlag)
      LUI_ButtonCreate("TOPO",    LUI_BUTTON_TOGGLE,0,2,BUTTONSIZE,0, topo_cb);
   else
      LUI_ButtonCreate("",    LUI_BUTTON_ONESHOT,0,2,BUTTONSIZE,0, NULL);
   if (MapFlag)
      LUI_ButtonCreate("MAP",     LUI_BUTTON_TOGGLE,1,2,BUTTONSIZE,0, map_cb);
   else
      LUI_ButtonCreate("",    LUI_BUTTON_ONESHOT,1,2,BUTTONSIZE,0, NULL);

   lb=LUI_ButtonCreate("BOX",  LUI_BUTTON_TOGGLE,2,2,BUTTONSIZE,0, box_cb);
   lb->state = LUI_BUTTON_IN;   LUI_ButtonDraw( lb );
   lb=LUI_ButtonCreate("CLOCK", LUI_BUTTON_TOGGLE,3,2,BUTTONSIZE,0, clock_cb);
   lb->state = LUI_BUTTON_IN;   LUI_ButtonDraw( lb );

   /** Row 4 **/
   LUI_ButtonCreate("SAVE",   LUI_BUTTON_ONESHOT,0,3,BUTTONSIZE,0, save_cb);
   LUI_ButtonCreate("RESTORE",LUI_BUTTON_ONESHOT,1,3,BUTTONSIZE,0, restore_cb);
   LUI_ButtonCreate("GRID #'s",LUI_BUTTON_TOGGLE,2,3,BUTTONSIZE,0, coord_cb);
   LUI_ButtonCreate("CONT #'s",LUI_BUTTON_TOGGLE,3,3,BUTTONSIZE,0, contnum_cb);

   /** Row 5 **/
   if (HQR_available)
     pretty_button = LUI_ButtonCreate("PRETTY", LUI_BUTTON_TOGGLE,
				      0,4,BUTTONSIZE,0, pretty_cb);
   else
     pretty_button = LUI_ButtonCreate("",     LUI_BUTTON_ONESHOT,
				      0,4,BUTTONSIZE,0, NULL );
   LUI_ButtonCreate("REVERSE",LUI_BUTTON_TOGGLE, 1,4,BUTTONSIZE,0, reverse_cb);
   LUI_ButtonCreate("INFO",   LUI_BUTTON_TOGGLE, 2,4,BUTTONSIZE,0, info_cb );
   LUI_ButtonCreate("",       LUI_BUTTON_ONESHOT,3,4,BUTTONSIZE,0, NULL);

   LUI_ButtonPadClose("Buttons");
   LUI_ButtonPadVisible("Buttons", 1);
   LUI_ButtonPadQuery("Buttons", &x, &y, &w, &h);
   CpHeight += h;


   /******************************************************************/
   /* Create the radio buttons.                                    ***/
   /******************************************************************/
   LUI_ButtonPadOpenType( "Modes",  /* name */
         LUI_BUTTON_RADIO,          /* type */
         CpWindow,                  /* parent window */
         BW, CpHeight,              /* position */
         BW,                        /* border size */
         0x000000                   /* background color */
       );
   lb = LUI_ButtonCreateType("Normal", 0,0, RADIOSIZE, MM_NORMAL, mode_cb);
   lb->state = LUI_BUTTON_IN;
   if (WindAvailable && NumTimes>1)
     LUI_ButtonCreateType( "Trajectory", 0,1, RADIOSIZE, MM_TRAJ,  mode_cb );
   else
     LUI_ButtonCreateType( " ", 0,1, RADIOSIZE, MM_TRAJ,  NULL );
   LUI_ButtonCreateType( "Slice",      0,2, RADIOSIZE, MM_SLICE, mode_cb );
   LUI_ButtonCreateType( "Label",      0,3, RADIOSIZE, MM_LABEL, mode_cb );
   LUI_ButtonPadClose( "Modes" );
   LUI_ButtonPadVisible( "Modes", 1 );
   LUI_ButtonPadQuery( "Modes", &x, &y, &w, &h );

   LUI_LabelOpen ( "ModeInfo",   /* name */
             6, ModeInfo1,       /* text strings */
             CpWindow,           /* parent window */
             BW+w, CpHeight,     /* origin relative to parent */
             CP_WIDTH-w, h,      /* dimensions */
             BW,                 /* border size */
             NULL                /* callback function */
             );
   LUI_LabelClose("ModeInfo");
   LUI_LabelVisible("ModeInfo", 1);
   LUI_LabelHighlight("ModeInfo", 1);
   LUI_LabelQuery("ModeInfo", &x, &y, &w, &h);
   CpHeight += h;


   /***********************************************************/
   /* Make wind slice buttons                                 */
   /***********************************************************/
   if (WindAvailable) {
      LUI_ButtonPadOpen ("WindButtons", /* name */
			 CpWindow,       /* parent window */
			 BW, CpHeight,   /* origin relative to parent */
			 BW,             /* border size */
			 0x000000        /* background color */
			 );

      LUI_ColorHilite = HWindColor[0];
      Button[NumButtons++] = LUI_ButtonCreate("H-WIND 1",
			       LUI_BUTTON_TOGGLE,0,0,BUTTONSIZE,0,hwind_cb);
      LUI_ColorHilite = VWindColor[0];
      Button[NumButtons++] = LUI_ButtonCreate("V-WIND 1",
                               LUI_BUTTON_TOGGLE,1,0,BUTTONSIZE,0,vwind_cb);
      LUI_ColorHilite = HWindColor[1];
      Button[NumButtons++] = LUI_ButtonCreate("H-WIND 2",
                               LUI_BUTTON_TOGGLE,2,0,BUTTONSIZE,1,hwind_cb);
      LUI_ColorHilite = VWindColor[1];
      Button[NumButtons++] = LUI_ButtonCreate("V-WIND 2",
                               LUI_BUTTON_TOGGLE,3,0,BUTTONSIZE,1,vwind_cb);

      LUI_ButtonPadClose("WindButtons");
      LUI_ButtonPadVisible("WindButtons", 1);
      LUI_ButtonPadQuery("WindButtons", &x, &y, &w, &h);
      CpHeight += h;
   }

   /************************************************************/
   /* Create contour and slice buttons.                        */
   /************************************************************/
   LUI_LabelOpen( "Columns",    /* name */
             2, Columns,        /* text strings */
             CpWindow,          /* parent window */
             BW, CpHeight,      /* origin relative to parent */
             CP_WIDTH, 40,      /* dimensions */
             BW,                /* border size */
             NULL               /* callback function */
             );
   LUI_LabelClose("Columns");
   LUI_LabelVisible("Columns", 1);
   LUI_LabelHighlight("Columns", 0);
   LUI_LabelQuery("Columns", &x, &y, &w, &h);
   CpHeight += h;

   LUI_ButtonPadOpen( "VarButtons",      /* name */
		      CpWindow,         /* parent window */
		      BW, CpHeight,     /* origin relative to parent */
		      BW,               /* border size */
		      0x000000          /* background color */
		      );

   /* Loop to create the parameters buttons */
   for (i=0; i<NumParms; i++) {
      for (j=0; j<5; j++) {
	 sprintf( label, "%s         %d", ParmName[i], j );
	 LUI_ColorHilite = Color[i][j];
	 Button[NumButtons++] = LUI_ButtonCreate(label, LUI_BUTTON_TOGGLE,
			  j,i,  /* position */
			  72,   /* width */
			  i*5+j, button_cb );
      }
   }

   LUI_ButtonPadClose("VarButtons");
   LUI_ButtonPadVisible("VarButtons", 1);
   LUI_ButtonPadQuery("VarButtons", &x, &y, &w, &h);
   CpHeight += h;


   /* Finish making the control panel. */
   CpHeight += BW;
   LUI_EventAdd ( CpWindow, ExposureMask | ButtonPressMask, main_process);
   LUI_MoveResizeWindow( CpWindow, cpx, cpy, CP_WIDTH, CpHeight );
   XMapWindow( wdpy, CpWindow );

   bottom = CpHeight + 2*SPACE;


   /************************************************************/
   /* Make wind slice density & scale window.                  */
   /************************************************************/
   if (WindAvailable) {
      WindWindow = LUI_CreateWindow( LUI_RootWindow, 0, 0 );

      LUI_LabelCreate( "WindLabel", WindLabStr, WindWindow,
		      BW, BW, 130,20,
		      BW, NULL, 0, 1 );
      LUI_LabelOpen( "WindScaleLabel", 1, WindScaleStr, WindWindow,
		    BW+130, BW, 80, 20,
		    BW, windscale_cb );
      LUI_LabelClose( "WindScaleLabel" );
      LUI_LabelVisible("WindScaleLabel", 1);
      LUI_LabelHighlight("WindScaleLabel", 0);
      LUI_LabelColor( "WindScaleLabel", 0xff00ffff );

      LUI_LabelCreate( "WindLabel2", WindLab2Str, WindWindow,
		      BW+210, BW, 80,20,
		      BW, NULL, 0, 1 );
      LUI_LabelOpen( "WindDensityLabel", 1, WindDensityStr, WindWindow,
		    BW+290, BW, 80, 20,
		    BW, winddensity_cb );
      LUI_LabelClose( "WindDensityLabel" );
      LUI_LabelVisible("WindDensityLabel", 1);
      LUI_LabelHighlight("WindDensityLabel", 0);
      LUI_LabelColor( "WindDensityLabel", 0xff00ffff );
      LUI_LabelQuery("WindDensityLabel", &x, &y, &w, &h);

      WindHeight = h + 2*BW;
      WindWidth = CP_WIDTH;

      /* Event Processing */
      LUI_EventAdd( WindWindow, ExposureMask | ButtonPressMask, main_process );
      LUI_MoveResizeWindow( WindWindow, cpx, bottom, WindWidth, WindHeight );
      XUnmapWindow( wdpy, WindWindow );

      bottom += WindHeight + SPACE;
   }

   /******************************************************************/
   /* create 3-D iso-level widget(s)                                 */
   /******************************************************************/

   IsoWindow = LUI_CreateWindow( LUI_RootWindow, cpx, cpy );

   LUI_SliderBoxOpen("IsoLevel", IsoWindow,  /* name, parent */
               BW, BW,       /* position */
               BW, NULL);    /* BW, func */
   IsoSlider = LUI_SliderCreate ( "dummy", 0.0, 1.0, 0.5,
			LUI_SLIDER_FLOAT , 0, 0, 0, 0,i, isosurface_cb);
   LUI_SliderBoxClose ("IsoLevel");
   LUI_SliderBoxQuery("IsoLevel", &x, &y, &w, &h);
   LUI_SliderBoxVisible("IsoLevel",1);

   IsosurfHeight = h + 2*BW;
   IsosurfWidth = w + 2*BW;

   /* Event Processing */
   LUI_EventAdd ( IsoWindow, ExposureMask | ButtonPressMask, main_process);
   LUI_MoveResizeWindow( IsoWindow, cpx, bottom, IsosurfWidth, IsosurfHeight );
   XUnmapWindow(wdpy,IsoWindow);

   bottom += IsosurfHeight + SPACE;


   /******************************************************************/
   /* create horizontal contour line interval widget                 */
   /******************************************************************/
   HSliceWindow = LUI_CreateWindow( LUI_RootWindow, 0, 0 );

   LUI_LabelCreate( "HSliceLabel", HSliceLabStr, HSliceWindow,
       BW, BW, 260,20,
       BW, NULL, 0, 1 );
   LUI_LabelOpen( "HSliceIntLabel", 1, HSliceIntStr, HSliceWindow,
       BW+260, BW, CP_WIDTH-260, 20,
       BW, hslice_cb );
   LUI_LabelClose( "HSliceIntLabel" );
   LUI_LabelVisible("HSliceIntLabel", 1);
   LUI_LabelHighlight("HSliceIntLabel", 0);
   LUI_LabelColor( "HSliceIntLabel", 0xff00ffff );
   LUI_LabelQuery("HSliceIntLabel", &x, &y, &w, &h);

   HSliceHeight = h + 2*BW;

   /* Event Processing */
   LUI_EventAdd( HSliceWindow, ExposureMask | ButtonPressMask, main_process );
   LUI_MoveResizeWindow( HSliceWindow, cpx, bottom, CP_WIDTH, HSliceHeight );
   XUnmapWindow( wdpy, HSliceWindow );
   bottom += HSliceHeight + SPACE;


   /******************************************************************/
   /* create vertical contour line interval widget                   */
   /******************************************************************/
   VSliceWindow = LUI_CreateWindow( LUI_RootWindow, 0, 0 );

   LUI_LabelCreate( "VSliceLabel", VSliceLabStr, VSliceWindow,
       BW, BW, 260,20,
       BW, NULL, 0, 1 );
   LUI_LabelOpen( "VSliceIntLabel", 1, VSliceIntStr, VSliceWindow,
       BW+260, BW, CP_WIDTH-260, 20,
       BW, vslice_cb );
   LUI_LabelClose( "VSliceIntLabel" );
   LUI_LabelVisible("VSliceIntLabel", 1);
   LUI_LabelHighlight("VSliceIntLabel", 0);
   LUI_LabelColor( "VSliceIntLabel", 0xff00ffff );
   LUI_LabelQuery("VSliceIntLabel", &x, &y, &w, &h);

   VSliceHeight = h + 2*BW;

   /* Event Processing */
   LUI_EventAdd( VSliceWindow, ExposureMask | ButtonPressMask, main_process );
   LUI_MoveResizeWindow( VSliceWindow, cpx, bottom, CP_WIDTH, VSliceHeight );
   XUnmapWindow( wdpy, VSliceWindow );
   bottom += VSliceHeight + SPACE;


   /******************************************************************/
   /* Create standard color selection window.                        */
   /******************************************************************/
   ColorWindow = LUI_CreateWindow( LUI_RootWindow, 100, 100 );

   LUI_SliderBoxOpen ("Color", ColorWindow,
       BW, BW, BW, NULL);
   red_slider = LUI_SliderCreate ( "Red", 0.0, 1.0, 0.5,
      LUI_SLIDER_FLOAT, 0,0, 0,0, 0, red_cb );
   green_slider = LUI_SliderCreate ( "Green", 0.0, 1.0, 0.5,
      LUI_SLIDER_FLOAT, 0,1, 0,0, 1, green_cb );
   blue_slider = LUI_SliderCreate ( "Blue", 0.0, 1.0, 0.5,
      LUI_SLIDER_FLOAT, 0,2, 0,0, 2, blue_cb );
   trans_slider = LUI_SliderCreate( "Transparency", 0.0, 1.0, 0.5,
      LUI_SLIDER_FLOAT, 0,3, 0,0, 3, trans_cb );
   LUI_SliderBoxClose ("Color");
   LUI_SliderBoxQuery("Color", &x, &y, &w, &h);
   LUI_SliderBoxVisible("Color",1);

   ColorHeight = h + 2*BW;
   ColorWidth = w + 2*BW;

   /* Event Processing */
   LUI_EventAdd ( ColorWindow, ExposureMask | ButtonPressMask, main_process );
   if (bottom+ColorHeight+SPACE>ScrHeight)
      bottom = ScrHeight - ColorHeight - SPACE;
   LUI_MoveResizeWindow( ColorWindow, cpx, bottom, ColorWidth, ColorHeight );
   XUnmapWindow( wdpy, ColorWindow );
   bottom += ColorHeight + SPACE;


   /****************************/
   /* Make Color widget window */
   /****************************/
   if (bottom+RGBHeight+SPACE>ScrHeight)
      bottom = ScrHeight - RGBHeight - SPACE;
   RGBWindow = create_cw( wdpy, ProgramName, cpx, bottom, RGBWidth, RGBHeight );

   for (i=0;i<MAXPARMS;i++) {
      int trans;
      trans = (Color[i][CHSLICE] & 0xff000000) >> 24;
      init_ctable( CHSliceColors[i], trans );
      CHSliceColors[i][255] = 0x00;  /* missing data -> transparent black */

      trans = (Color[i][CVSLICE] & 0xff000000) >> 24;
      init_ctable( CVSliceColors[i], trans );
      CVSliceColors[i][255] = 0x00;  /* missing data -> transparent black */
   }

   bottom += RGBHeight + SPACE;


   /**************************/
   /* Make Trajectory Window */
   /**************************/
   TrajWindow = LUI_CreateWindow(LUI_RootWindow, 100, 100 );
   TrajHeight = BW;

   /******************************************************************/
   /* Create the user trajectories title label.                      */
   /******************************************************************/
   LUI_LabelOpen ( "TitleTraj",      /* name */
		   2, TitleTraj,     /* text strings */
                   TrajWindow,       /* parent window */
		   BW, TrajHeight,   /* origin relative to parent */
		   CP_WIDTH, 36,     /* dimensions */
		   BW,               /* border size */
		   NULL              /* callback function */
	          );
   LUI_LabelClose("TitleTraj");
   LUI_LabelVisible("TitleTraj", 1);   
   LUI_LabelHighlight("TitleTraj", 0);
   LUI_LabelQuery("TitleTraj", &x, &y, &w, &h);
   TrajHeight += h;

   /* Trajectory set buttons */
   LUI_ButtonPadOpen ("TrajButtons",   /* name */
                 TrajWindow,           /* parent window */
		 BW, TrajHeight,       /* origin relative to parent */
		 BW,                   /* border size */
		 0x000000              /* background color */	
		 );
   for (i=0;i<TRAJGROUPS;i++) {
      sprintf(label, "Set %d", i+1 );
      LUI_ColorHilite = TrajColor[i];
      Button[NumButtons++] =
	LUI_ButtonCreate(label,LUI_BUTTON_TOGGLE,i%4,i/4, 72, i, traj_cb);
      if (DisplayTraj[i]) {
	 Button[NumButtons-1]->state = LUI_BUTTON_IN;
	 LUI_ButtonDraw( Button[NumButtons-1] );
      }
   }
   LUI_ButtonPadClose("TrajButtons");
   LUI_ButtonPadVisible("TrajButtons", 1);
   LUI_ButtonPadQuery("TrajButtons", &x, &y, &w, &h);
   TrajHeight += h;

   /* trajectory type-ins for step and length */
   LUI_LabelCreate( "TrajStepLabel", TrajStepLabel, TrajWindow,
       BW, TrajHeight, 70,20,
       BW, NULL, 0, 1 );
   LUI_LabelOpen( "TrajStepValue", 1, TrajStepValue, TrajWindow,
       BW+70, TrajHeight, 80,20,
       BW, trajstep_cb );
   LUI_LabelClose( "TrajStepValue" );
   LUI_LabelVisible("TrajStepValue", 1);
   LUI_LabelHighlight("TrajStepValue", 0);
   LUI_LabelColor( "TrajStepValue", 0xff00ffff );

   LUI_LabelCreate( "TrajLengthLabel", TrajLengthLabel, TrajWindow,
       BW+150, TrajHeight, 70,20,
       BW, NULL, 0, 1 );
   LUI_LabelOpen( "TrajLengthValue", 1, TrajLengthValue, TrajWindow,
       BW+220, TrajHeight, 80,20,
       BW, trajlength_cb );
   LUI_LabelClose( "TrajLengthValue" );
   LUI_LabelVisible("TrajLengthValue", 1);
   LUI_LabelHighlight("TrajLengthValue", 0);
   LUI_LabelColor( "TrajLengthValue", 0xff00ffff );
   LUI_LabelQuery("TrajLengthValue", &x, &y, &w, &h);

   TrajHeight += h;


   /* Trajectory control buttons */
   LUI_ButtonPadOpen ("TrajButtons2",   /* name */
                 TrajWindow,           /* parent window */
		 BW, TrajHeight,       /* origin relative to parent */
		 BW,                   /* border size */
		 0x000000              /* background color */	
		 );
   LUI_ButtonCreate("Ribbon",      LUI_BUTTON_TOGGLE, 0,0, 96, 0, ribbon_cb);
   LUI_ButtonCreate("Delete Last", LUI_BUTTON_ONESHOT,1,0, 96, 0, dellast_cb);
   LUI_ButtonCreate("Delete Set",  LUI_BUTTON_ONESHOT,2,0, 96, 0, delset_cb);
   LUI_ButtonPadClose("TrajButtons2");
   LUI_ButtonPadVisible("TrajButtons2", 1);
   LUI_ButtonPadQuery("TrajButtons2", &x, &y, &w, &h);
   TrajHeight += h + BW;
   TrajWidth = 2*BW+w;

   LUI_EventAdd ( TrajWindow, ExposureMask | ButtonPressMask, main_process);
   LUI_MoveResizeWindow( TrajWindow, 400,875, TrajWidth,TrajHeight );
   XUnmapWindow(wdpy, TrajWindow );


   /**********************************************************************/
   /*  Create the verify window                                          */
   /**********************************************************************/
   VerifyWindow = LUI_CreateWindow( LUI_RootWindow, 100, 100 );

   LUI_LabelOpen ( "VerifyLabel",    /* name */
             1, VerifyMessage,       /* text strings */
             VerifyWindow,           /* parent window */
             BW, BW,                 /* origin relative to parent */
             CP_WIDTH-2*BW, 50,      /* size */
             BW,                     /* border size */
             NULL                    /* callback function */
             );
   LUI_LabelClose("VerifyLabel");
   LUI_LabelVisible("VerifyLabel", 1);
   LUI_LabelHighlight("VerifyLabel", 0);

   LUI_ButtonPadOpen ("Verify",  /* name */
		 VerifyWindow,   /* parent window */
		 BW, 50+BW,      /* origin relative to parent */
		 BW,             /* border size */
		 0x000000
		 );
   LUI_ButtonCreate("OK",    LUI_BUTTON_ONESHOT, 0,0, 180, NULL, ok_cb);
   LUI_ButtonCreate("CANCEL",LUI_BUTTON_ONESHOT, 1,0, 180, NULL, cancel_cb);
   LUI_ButtonPadClose("Verify");
   LUI_ButtonPadVisible("Verify", 1);
   LUI_ButtonPadQuery("Verify", &x, &y, &w, &h);
   VerifyHeight = BW+50+h+BW;

   /* Event Processing */
   LUI_MoveResizeWindow( VerifyWindow, 600,400, CP_WIDTH,VerifyHeight );
   LUI_EventAdd( VerifyWindow, ExposureMask | ButtonPressMask, main_process);
   XUnmapWindow( wdpy, VerifyWindow );


   /**********************************************************************/
   /*  Create the alert window                                          */
   /**********************************************************************/
   AlertWindow = LUI_CreateWindow( LUI_RootWindow, 100, 100 );

   LUI_LabelOpen ( "AlertLabel",    /* name */
             2, AlertMessage,       /* text strings */
             AlertWindow,           /* parent window */
             BW, BW,                /* origin relative to parent */
             CP_WIDTH-2*BW, 50,     /* size */
             BW,                    /* border size */
             NULL                   /* callback function */
             );
   LUI_LabelClose("AlertLabel");
   LUI_LabelVisible("AlertLabel", 1);
   LUI_LabelHighlight("AlertLabel", 0);

   LUI_ButtonPadOpen ("Alert",  /* name */
		 AlertWindow,   /* parent window */
		 BW, 50+BW,     /* origin relative to parent */
		 BW,            /* border size */
		 0x000000
		 );
   LUI_ButtonCreate("          OK", LUI_BUTTON_ONESHOT, 0,0, 360, NULL, ok_cb);
   LUI_ButtonPadClose("Alert");
   LUI_ButtonPadVisible("Alert", 1);
   LUI_ButtonPadQuery("Alert", &x, &y, &w, &h);
   AlertHeight = BW+50+h+BW;

   /* Event Processing */
   LUI_MoveResizeWindow( AlertWindow, 600,400, CP_WIDTH,AlertHeight );
   LUI_EventAdd( AlertWindow, ExposureMask | ButtonPressMask, main_process);
   XUnmapWindow( wdpy, AlertWindow );

   /**********************************************************************/
   /*  Create the clone window                                           */
   /**********************************************************************/
   CloneWindow = LUI_CreateWindow( LUI_RootWindow, 100, 100 );

   LUI_LabelOpen( "CloneTitle", /* name */
             1, CloneTitle,     /* text strings */
             CloneWindow,       /* parent window */
             BW, BW,            /* origin relative to parent */
             RADIOSIZE*2+BW*2, 25,  /* dimensions */
             BW,                    /* border size */
             NULL                   /* callback function */
             );
   LUI_LabelClose("CloneTitle");
   LUI_LabelVisible("CloneTitle", 1);
   LUI_LabelHighlight("CloneTitle", 0);
   LUI_LabelQuery("CloneTitle", &x, &y, &w, &h);
   CloneHeight = BW+h;

#ifdef OLD
   LUI_ButtonPadOpenType( "Clone",  /* name */
         LUI_BUTTON_RADIO,          /* type */
         CloneWindow,               /* parent window */
         BW, CloneHeight,           /* position in window */
         BW,                        /* border size */
         0x000000                   /* background color */
       );

   j = (NumParms+1) / 2;
   for (i=0;i<NumParms;i++) {
      x = (i<j) ? 0 : 1;
      y = (i<j) ? i : i-j;
      LUI_ButtonCreateType( ParmName[i], x,y, RADIOSIZE, i, do_clone_cb );
   }
   LUI_ButtonPadClose( "Clone" );
   LUI_ButtonPadVisible( "Clone", 1 );
   LUI_ButtonPadQuery( "Clone", &x, &y, &w, &h );
   CloneWidth = w+2*BW;
   CloneHeight += h;

   LUI_ButtonPadOpen ("CloneButtons",  /* name */
		 CloneWindow,          /* parent window */
		 BW, CloneHeight,      /* origin relative to parent */
		 BW,                   /* border size */
		 0x000000
		 );

   LUI_ButtonCreate("OK",    LUI_BUTTON_ONESHOT,0,0,RADIOSIZE,NULL,ok_cb);
   LUI_ButtonCreate("CANCEL",LUI_BUTTON_ONESHOT,1,0,RADIOSIZE,NULL,cancel_cb);
   LUI_ButtonPadClose("CloneButtons");
   LUI_ButtonPadVisible("CloneButtons", 1);
   LUI_ButtonPadQuery("CloneButtons", &x, &y, &w, &h);
   CloneHeight += h;
#else
   LUI_ButtonPadOpen( "Clone",  /* name */
         CloneWindow,           /* parent window */
         BW, CloneHeight,       /* position in window */
         BW,                    /* border size */
         0x000000               /* background color */
       );

   j = (NumParms+1) / 2;
   for (i=0;i<NumParms;i++) {
      x = (i<j) ? 0 : 1;
      y = (i<j) ? i : i-j;
      LUI_ButtonCreate( ParmName[i], LUI_BUTTON_ONESHOT,
		       x,y, RADIOSIZE, i, do_clone_cb );
   }
   LUI_ButtonPadClose( "Clone" );
   LUI_ButtonPadVisible( "Clone", 1 );
   LUI_ButtonPadQuery( "Clone", &x, &y, &w, &h );
   CloneWidth = w+2*BW;
   CloneHeight += h;

   LUI_ButtonPadOpen ("CloneButtons",  /* name */
		 CloneWindow,          /* parent window */
		 BW, CloneHeight,      /* origin relative to parent */
		 BW,                   /* border size */
		 0x000000
		 );

   LUI_ButtonCreate("        CANCEL",LUI_BUTTON_ONESHOT, 0,0,
		    2*RADIOSIZE, -1 ,do_clone_cb);
   LUI_ButtonPadClose("CloneButtons");
   LUI_ButtonPadVisible("CloneButtons", 1);
   LUI_ButtonPadQuery("CloneButtons", &x, &y, &w, &h);
   CloneHeight += h;
#endif


   CloneHeight += BW;

   LUI_EventAdd ( CloneWindow, ExposureMask | ButtonPressMask, main_process);
   LUI_MoveResizeWindow( CloneWindow, 450,400, CloneWidth, CloneHeight );
   XUnmapWindow( wdpy, CloneWindow );

}



/*** reset_widgets ****************************************************
   Reset the widgets after a RESTORE by updating colors, isolevels,
   intervals, etc.
**********************************************************************/
static void reset_widgets()
{
   int i, j;
   unsigned long *color;

   if (ColorPtr) {
      XUnmapWindow( wdpy, ColorWindow );
      color = ColorPtr;
      ColorPtr = NULL;
      map_color_window( color );
   }

   if (ip3d>=0) {
      XUnmapWindow( wdpy, IsoWindow );
      i = ip3d;
      ip3d = -1;
      map_isosurf_window( i );
   }

   if (ip2d>=0) {
      XUnmapWindow( wdpy, HSliceWindow );
      i = ip2d;
      ip2d = -1;
      map_hslice_window( i );
   }

   if (ipxs>=0) {
      XUnmapWindow( wdpy, VSliceWindow );
      i = ipxs;
      ipxs = -1;
      map_vslice_window( i );
   }

   if (iphs>=0 || ipvs>=0 || iptopo) {
      XUnmapWindow( wdpy, RGBWindow );
      if (iphs>=0)
	map_chslice_window( iphs );
      else if (ipvs>=0)
	map_cvslice_window( ipvs );
      else
	map_topo_color_window();
   }

   if (iphw>=0 || ipvw>=0) {
      XUnmapWindow( wdpy, WindWindow );
      if (iphw>=0)
	map_hwind_window( iphw );
      else
	map_vwind_window( ipvw );
   }

   /* redraw all buttons with restored colors */
   for (i=0;i<NumButtons;i++) {
      if (Button[i]->func==button_cb) {
	 /* variable button */
	 int row, col;
	 row = Button[i]->index / 5;
	 col = Button[i]->index % 5;
	 LUI_ColorHilite = Color[row][col];
      }
      else if (Button[i]->func==traj_cb) {
	 LUI_ColorHilite = TrajColor[ Button[i]->index ];
      }
      else if (Button[i]->func==hwind_cb) {
	 LUI_ColorHilite = HWindColor[ Button[i]->index ];
      }
      else if (Button[i]->func==vwind_cb) {
	 LUI_ColorHilite = VWindColor[ Button[i]->index ];
      }
      LUI_ButtonDraw( Button[i] );
   }

}



/* This is called by LUI when it want's to pass us an event */
int main_process( window, event )
Window window;
XEvent *event;
{

   switch (event->type) {
   case Expose:
      if (event->xexpose.count==0) {
	 if (event->xexpose.window == CpWindow)
            LUI_FrameWindow( window, CP_WIDTH,CpHeight, BW );
	 else if (event->xexpose.window == HSliceWindow)
            LUI_FrameWindow( window, CP_WIDTH,HSliceHeight, BW );
	 else if (event->xexpose.window == VSliceWindow)
            LUI_FrameWindow( window, CP_WIDTH,VSliceHeight, BW );
	 else if (event->xexpose.window == ColorWindow)
            LUI_FrameWindow( window, ColorWidth,ColorHeight, BW );
	 else if (event->xexpose.window == TrajWindow)
            LUI_FrameWindow( window, TrajWidth,TrajHeight, BW );
	 else if (event->xexpose.window == VerifyWindow)
            LUI_FrameWindow( window, CP_WIDTH,VerifyHeight, BW );
	 else if (event->xexpose.window == AlertWindow)
            LUI_FrameWindow( window, CP_WIDTH,AlertHeight, BW );
	 else if (event->xexpose.window == CloneWindow)
            LUI_FrameWindow( window, CloneWidth,CloneHeight, BW );
	 else if (event->xexpose.window == WindWindow)
            LUI_FrameWindow( window, WindWidth,WindHeight, BW );
	 else if (event->xexpose.window == RGBWindow) {
	    int x, y, bw, d;
	    Window r;
	    XGetGeometry( wdpy, window, &r, &x, &y,
			 &RGBWidth, &RGBHeight, &bw, &d );
            LUI_FrameWindow( window, RGBWidth,RGBHeight, BW );
	 }
	 else if (event->xexpose.window == IsoWindow) {
	    LUI_FrameWindow( window, IsosurfWidth, IsosurfHeight, BW );
	 }

      }
      XMapWindow( wdpy, window );
      break;
   }

}




/*** set_contour_range ************************************************
   Set the range of values for the contour slider.
   Input:  min - the minimum value.
           max - the maximum value.
           val - the current value to use.
   Output:  none.
**********************************************************************/
void set_contour_range( min, max, val )
float min, max, val;
{
   LUI_SLIDER  *s;

   s = LUI_SliderLookup("Contour");

   if (s) {
      s->low = min;
      s->high = max;
      s->value = val;
      s->old_value = val;
      LUI_SliderDraw( s );
   }
}





/*******************************************************/
/*** CALL BACK FUNCTIONS FOR VERIFY AND ALERT WINDOW ***/
/*******************************************************/

static int verify_value;

static ok_cb()
{
  verify_value = 1;
}


static cancel_cb()
{
  verify_value = 0;
}


/*** verify ***********************************************************
   Open the verification window with a prompt asking the user to
   specify either an OK or CANCEL.
   Input:  prompt - a string which asks the user what he wants to do.
                    Ex: "Do you want to exit?"
   Returned:  1 = OK selected
              0 = CANCEL selected
**********************************************************************/
int verify( prompt )
char *prompt;
{
   XEvent pe;

   strcpy( VerifyMessage[0], prompt );

   LUI_LabelChangeText("VerifyLabel", 1, VerifyMessage);

   verify_value = -1;
   XMapWindow( wdpy, VerifyWindow );

   while (verify_value < 0) {
      if (XPending(wdpy)) {
         XNextEvent(wdpy, &pe);
         LUI_EventDispatch(&pe);
      }
   }
   XUnmapWindow( wdpy, VerifyWindow );
   Redraw = 1;
   return(verify_value);
}



/*** alert ***********************************************************
   Open the alert window with a message and wait until the user clicks
   on "OK".
   Input:  prompt - a message string
**********************************************************************/
int alert( prompt1, prompt2 )
char *prompt1, *prompt2;
{
   XEvent pe;

   strcpy( AlertMessage[0], prompt1 );
   strcpy( AlertMessage[1], prompt2 );

   LUI_LabelChangeText("AlertLabel", 2, AlertMessage);

   verify_value = -1;
   XMapWindow( wdpy, AlertWindow );

   while (verify_value < 0) {
      if (XPending(wdpy)) {
         XNextEvent(wdpy, &pe);
         LUI_EventDispatch(&pe);
      }
   }
   XUnmapWindow( wdpy, AlertWindow );
   Redraw = 1;
   return 1;
}



debug_cb()
{
   Debug = !Debug;
}



/****************************************************/
/*** CALL-BACK FUNCTIONS FOR MAIN CONTROL BUTTONS ***/
/****************************************************/


/*** anim_cb **********************************************************
   Animate button call-back function.  Toggle animation on/off.
   Input:  pb - LUI button pointer
**********************************************************************/
static int anim_cb( pb )
LUI_BUTTON *pb;
{
   int i, time, p;

   if (GoTime)
      GoTime = 0;
   else {
      if (pb->event->xbutton.button == Button3)
	 /* reverse animation */
 	 GoTime = -1;
      else
	 /* forward animation */
 	 GoTime = 1;
   }

   Redraw = 2;
}




/*** step_cb **********************************************************
   Step button call-back function.  Step forward or backward one time
   step or set time to first time step depending on which mouse
   button was pressed.
   Input:  pb - LUI button pointer
**********************************************************************/
static int step_cb(pb)
LUI_BUTTON *pb;
{
   int ip, w;

   /* CurTime must be in [0..NumTimes-1] */

   if (GoTime == 0) {
      if (pb->event->xbutton.button == Button1) {
         if (CurTime==NumTimes-1)
            CurTime = 0;
         else
            CurTime++;
      }
      else if(pb->event->xbutton.button == Button2) {
         CurTime = 0;
      }
      else if(pb->event->xbutton.button == Button3) {
         if (CurTime==0)
            CurTime = NumTimes-1;
         else
            CurTime--;
      }

      /* Redraw display */
      Redraw = 1;
   }

}



/* callback for quit button */
static int exit_cb()
{
   if (verify("Do you really want to exit?")) {
      /* send quit message to server */
      request_quit();

      terminate();
   }
}



static int box_cb()
{
   /* toggle box rendering */
   DisplayBox = !DisplayBox;
   Redraw = 1;
}


static int clock_cb()
{
   /* toggle clock rendering */
   DisplayClock = !DisplayClock;
   Redraw = 1;
}



static int save_cb( pb )
LUI_BUTTON *pb;
{
   char savefile[100], str[100], prompt[100];
   int cflag, res;

   if (pb->event->xbutton.button==Button3) {
      /* right mouse button -> just save colors */
      strcpy( prompt, "Save colors only?" );
      cflag = 1;
   }
   else {
      /* left / middle button -> save everything */
      strcpy( prompt, "Save graphics and colors?" );
      cflag = 0;
   }

   if (verify(prompt)) {
      sprintf( savefile, "%s.SAVE", DataFile );
      res = save(savefile,cflag);
      if (res==0) {
	 /* printf("Saved: %s\n", savefile ); */
	 Redraw = 1;
      }
      else if (res==1) {
	 sprintf( str, "%s for writing", savefile );
	 alert("Error:  unable to open", str );
      }
      else if (res==2) {
	 sprintf( str, "Error while writing %s", savefile );
	 alert( str, "Check if disk is full." );
      }
   }
}


static int restore_cb()
{
   char savefile[100], str[100];
   int res;

   if (verify("Restore saved graphics/colors?")) {
      sprintf( savefile, "%s.SAVE", DataFile );
      res = restore(savefile);
      if (res==0) {
	 /*printf("Restored: %s\n", savefile );*/
	 reset_widgets();
	 Redraw = 1;
      }
      else if (res==1) {
	 sprintf( str, "%s for reading", savefile );
	 alert("Error:  unable to open", str );
      }
      else if (res==2) {
	 sprintf( str, "Error while reading %s", savefile );
	 alert( str, "File may be corrupt or an old version" );
      }
   }
}



/* callback for topography button */
static int topo_cb( pb )
LUI_BUTTON *pb;
{
   if (pb->event->xbutton.button==Button1) {
      /* toggle display of topography */
      DisplayTopo = !DisplayTopo;
      Redraw = 1;
   }
   else if (pb->event->xbutton.button==Button2 ||
	    pb->event->xbutton.button==Button3) {
      /* topo color widget */
      if (pb->state == LUI_BUTTON_OUT)
         pb->state = LUI_BUTTON_IN;
      else
         pb->state = LUI_BUTTON_OUT;
      if (iptopo)
	XUnmapWindow( wdpy, RGBWindow );
      else {
	 iphs = ipvs = -1;
	 map_topo_color_window();
      }
      iptopo = !iptopo;
   }
}



/* callback for map display toggle */
static int map_cb()
{
   DisplayMap = !DisplayMap;
   Redraw = 1;
}




/* toggle perspective/parallel projection */
static int proj_cb()
{
   Perspec = !Perspec;
   Redraw = 1;
}



/* toggle contour label drawing */
static int contnum_cb()
{
   ContnumFlag = !ContnumFlag;
   Redraw = 1;
}



/* toggle between geographic and grid coordinate display */
static int coord_cb()
{
   CoordFlag = !CoordFlag;
   Redraw = 1;
}


static int hwind_cb( pb )
LUI_BUTTON *pb;
{
   int i;

   i = pb->index;

   if (pb->event->xbutton.button==Button1) {
      /* toggle display on/off */
      DisplayHWind[i] = !DisplayHWind[i];
      LUI_ColorHilite = HWindColor[i];
      if (DisplayHWind[i]) {
	 iphw = i;
	 map_hwind_window( iphw );
	 request_hwindslice( CurTime, i );
      }
      else if (iphw>=0) {
	 XUnmapWindow(wdpy,WindWindow);
	 iphw = -1;
      }
      ipvw = -1;
   }
   else if (pb->event->xbutton.button==Button2) {
      /* toggle widget on/off */
      if (pb->state == LUI_BUTTON_OUT)
         pb->state = LUI_BUTTON_IN;
      else
         pb->state = LUI_BUTTON_OUT;
      LUI_ColorHilite = HWindColor[i];
      if (iphw==i) {
	 XUnmapWindow(wdpy,WindWindow);
	 iphw = -1;
      }
      else {
	 iphw = i;
	 map_hwind_window( iphw );
      }
      ipvw = -1;
   }
   else if (pb->event->xbutton.button==Button3) {
      /* change color */
      if (pb->state == LUI_BUTTON_OUT)
         pb->state = LUI_BUTTON_IN;
      else
         pb->state = LUI_BUTTON_OUT;
      LUI_ColorHilite = HWindColor[i];
      Last_button = pb;
      map_color_window( &HWindColor[i] );
   }
   Redraw = 1;
}


static int vwind_cb( pb )
LUI_BUTTON *pb;
{
   int i;

   i = pb->index;

   if (pb->event->xbutton.button==Button1) {
      /* toggle display on/off */
      DisplayVWind[i] = !DisplayVWind[i];
      LUI_ColorHilite = VWindColor[i];
      if (DisplayVWind[i]) {
	 ipvw = i;
	 map_vwind_window( ipvw );
	 request_vwindslice( CurTime, i );
      }
      else if (ipvw>=0) {
	 XUnmapWindow(wdpy,WindWindow);
	 ipvw = -1;
      }
      iphw = -1;
   }
   else if (pb->event->xbutton.button==Button2) {
      /* toggle widget on/off */
      if (pb->state == LUI_BUTTON_OUT)
         pb->state = LUI_BUTTON_IN;
      else
         pb->state = LUI_BUTTON_OUT;
      LUI_ColorHilite = VWindColor[i];
      if (ipvw==i) {
	 XUnmapWindow(wdpy,WindWindow);
	 ipvw = -1;
      }
      else {
	 ipvw = i;
	 map_vwind_window( ipvw );
      }
      iphw = -1;
   }
   else if (pb->event->xbutton.button==Button3) {
      /* change color */
      if (pb->state == LUI_BUTTON_OUT)
         pb->state = LUI_BUTTON_IN;
      else
         pb->state = LUI_BUTTON_OUT;
      LUI_ColorHilite = VWindColor[i];
      Last_button = pb;
      map_color_window( &VWindColor[i] );
   }
   Redraw = 1;
}




/*** view_cb **********************************************************
   Set the view to either North, South, East, West, Top, or Bottom.
**********************************************************************/
static view_cb( b )
LUI_BUTTON *b;
{
   if (b->event->xbutton.button==Button3) {
      if (b->index==1)
	 /* bottom */
         make_matrix( 0.0, 180.0, 0.0, DEFAULT_SCALE, 0.0, 0.0, 0.0, CTM );
      else if (b->index==2)
	 /* north */
         make_matrix( 90.0, 180.0, 0.0, DEFAULT_SCALE, 0.0, 0.0, 0.0, CTM );
      else
	 /* east */
         make_matrix( 90.0, 90.0, 0.0, DEFAULT_SCALE, 0.0, 0.0, 0.0, CTM );
   }
   else {
      if (b->index==1)
	 /* top */
         make_matrix( 0.0, 0.0, 0.0, DEFAULT_SCALE, 0.0, 0.0, 0.0, CTM );
      else if (b->index==2)
	 /* south */
         make_matrix( 90.0, 0.0, 0.0, DEFAULT_SCALE, 0.0, 0.0, 0.0, CTM );
      else
	 /* west */
         make_matrix( 90.0, -90.0, 0.0, DEFAULT_SCALE, 0.0, 0.0, 0.0, CTM );
   }

   mat_inv( CTMinv, CTM );
   Redraw = 1;
}




/*** mode_cb ******************************************************
   This function is called when the user clicks on any of the mode
   radio button widgets.
**********************************************************************/
static mode_cb( pb )
LUI_BUTTON *pb;
{

   if (MouseMode==MM_TRAJ)
      XUnmapWindow( wdpy, TrajWindow );

   if (MouseMode==MM_LABEL)
      finish_making_label();

   MouseMode = pb->index;

   switch (MouseMode) {
      case MM_NORMAL:
	 LUI_LabelChangeText( "ModeInfo", 6, ModeInfo1 );
	 DisplayCursor = 0;
	 break;
      case MM_TRAJ:
	 LUI_LabelChangeText( "ModeInfo", 6, ModeInfo2 );
	 XMapWindow( wdpy, TrajWindow );
	 DisplayCursor = 1;
	 break;
      case MM_LABEL:
	 LUI_LabelChangeText( "ModeInfo", 6, ModeInfo3 );
	 DisplayCursor = 0;
	 break;
      case MM_SLICE:
	 LUI_LabelChangeText( "ModeInfo", 6, ModeInfo4 );
	 DisplayCursor = 0;
	 break;
   }

   Redraw = 1;
}



/* trajectory set buttons */
static traj_cb(pb)
LUI_BUTTON *pb;
{
   int i;

   i = pb->index;

   if (pb->event->xbutton.button==Button1) {
      /*** Left Button -> toggle set on/off ***/
      DisplayTraj[i] = !DisplayTraj[i];
      LUI_ColorHilite = TrajColor[i];
      if (DisplayTraj[i]) {
	 CursorColor = &TrajColor[i];
	 iptraj = i;
      }
   }
   else if (pb->event->xbutton.button==Button2){
      /* Middle Button -> select set */
      if (pb->state == LUI_BUTTON_OUT)
         pb->state = LUI_BUTTON_IN;
      else
         pb->state = LUI_BUTTON_OUT;
      LUI_ColorHilite = TrajColor[i];
      CursorColor = &TrajColor[i];
      iptraj = i;
   }
   else if (pb->event->xbutton.button==Button3) {
      /* Right Button -> color selector */
      if (pb->state == LUI_BUTTON_OUT)
         pb->state = LUI_BUTTON_IN;
      else
         pb->state = LUI_BUTTON_OUT;
      LUI_ColorHilite = TrajColor[i];
      Last_button = pb;
      map_color_window( &TrajColor[i] );
   }

   Redraw = 1;
}


/* toggle trajectory ribbon button */
static ribbon_cb()
{
   RibbonFlag = !RibbonFlag;
   Redraw = 1;
}


/* delete last trajectory */
static dellast_cb()
{
   del_last_traj();
   Redraw = 1;
}


/* delete trajectory set */
static delset_cb()
{
   char msg[100];

   sprintf(msg,"Delete trajectory set %d?", iptraj+1);
   if (verify(msg)) {
      del_traj_group( iptraj );
   }
   Redraw = 1;
}


/* trajectory step type-in */
static trajstep_cb( lab )
LUI_LABEL *lab;
{
   int trajstep_cb2();

   if (lab->event->type==KeyPress) {
      LUI_LabelUserInput( lab, 0, trajstep_cb2, lab->event, 1 );
      XSync( wdpy, 0 );
   }
   else if (lab->event->type==EnterNotify) {
      /* clear the type-in */
      strcpy( TrajStepValue[0], "" );
      LUI_LabelChangeText( "TrajStepValue", 1, TrajStepValue );
      LUI_LabelHighlight( "TrajStepValue", 1 );
      XSync( wdpy, 0 );
   }
   else if (lab->event->type==LeaveNotify) {
      /* restore the type-in */
      sprintf( TrajStepValue[0], "%6.2f", TrajStep );
      LUI_LabelChangeText( "TrajStepValue", 1, TrajStepValue );
      LUI_LabelHighlight( "TrajStepValue", 0 );
      XSync( wdpy, 0 );
   }
}


static trajstep_cb2( valstr )
char *valstr;
{
   sscanf( valstr, "%f", &TrajStep );  /* convert from ascii to float */
   if (TrajStep<=0.0)
      TrajStep = 1.0;
}


/* trajectory length type-in */
static trajlength_cb( lab )
LUI_LABEL *lab;
{
   int trajlength_cb2();

   if (lab->event->type==KeyPress) {
      LUI_LabelUserInput( lab, 0, trajlength_cb2, lab->event, 1 );
      XSync( wdpy, 0 );
   }
   else if (lab->event->type==EnterNotify) {
      /* clear the type-in */
      strcpy( TrajLengthValue[0], "" );
      LUI_LabelChangeText( "TrajLengthValue", 1, TrajLengthValue );
      LUI_LabelHighlight( "TrajLengthValue", 1 );
      XSync( wdpy, 0 );
   }
   else if (lab->event->type==LeaveNotify) {
      /* restore the type-in */
      sprintf( TrajLengthValue[0], "%6.2f", TrajLength );
      LUI_LabelChangeText( "TrajLengthValue", 1, TrajLengthValue );
      LUI_LabelHighlight( "TrajLengthValue", 0 );
      XSync( wdpy, 0 );
   }
}


static trajlength_cb2( valstr )
char *valstr;
{
   sscanf( valstr, "%f", &TrajLength );  /* convert from ascii to float */
   if (TrajLength<=0.0)
     TrajLength = 1.0;
}




/**********************************************************/
/*** CALL-BACK FUNCTIONS FOR MATRIX OF GRAPHICS BUTTONS ***/
/**********************************************************/


static button_cb(pb)
LUI_BUTTON *pb;
{
   int index, row, col;

   index = pb->index;
   row = index / 5;
   col = index % 5;

   if (pb->event->xbutton.button==Button1) {
      /*** Left Button -> toggle graphic on/off ***/
      LUI_ColorHilite = Color[row][col];
      switch (col) {
      case ISOSURF:
	 /** Iso-Surface **/
	 DisplaySurf[row] = !DisplaySurf[row];
	 if (ip3d==row) {
	    XUnmapWindow(wdpy,IsoWindow);
	    ip3d = -1;
	 }
	 if (DisplaySurf[row]) {
	    if (ip3d>=0) {
	       XUnmapWindow(wdpy,IsoWindow);
	    }
	    ip3d = row;
	    map_isosurf_window( ip3d );
	 }
         break;
      case HSLICE:
	 /** Horizontal Contour Line Slice **/
	 DisplayHSlice[row] = !DisplayHSlice[row];
	 if (ip2d==row) {
	    XUnmapWindow(wdpy,HSliceWindow);
	    ip2d = -1;
	 }
	 if (DisplayHSlice[row]) {
	    if (ip2d>=0)
	       XUnmapWindow(wdpy,HSliceWindow);
	    ip2d = row;
	    map_hslice_window( ip2d );
	    request_hslice( CurTime, row );
         }
	 break;
      case VSLICE:
	 /** Vertical Contour Line Slice **/
	 DisplayVSlice[row] = !DisplayVSlice[row];
	 if (ipxs==row) {
	    XUnmapWindow(wdpy,VSliceWindow);
	    ipxs = -1;
	 }
	 if (DisplayVSlice[row]) {
	    if (ipxs>=0)
	       XUnmapWindow(wdpy,VSliceWindow);
	    ipxs = row;
	    map_vslice_window( ipxs );
	    request_vslice( CurTime, row );
         }
	 break;
      case CHSLICE:
	 /** Horizontal Colored Slice **/
	 DisplayCHSlice[row] = !DisplayCHSlice[row];
	 if (DisplayCHSlice[row]) {
	    iphs = row;
	    map_chslice_window( iphs );
	    request_chslice( CurTime, row );
	 }
	 else if (iphs>=0) {
	    XUnmapWindow(wdpy,RGBWindow);
	    iphs = -1;
	 }
	 ipvs = -1;
	 iptopo = 0;
	 break;
      case CVSLICE:
	 /** Vertical Colored Slice **/
	 DisplayCVSlice[row] = !DisplayCVSlice[row];
	 if (DisplayCVSlice[row]) {
	    ipvs = row;
	    map_cvslice_window( ipvs );
	    request_cvslice( CurTime, row );
	 }
	 else if (ipvs>=0) {
	    XUnmapWindow(wdpy,RGBWindow);
	    ipvs = -1;
	 }
	 iphs = -1;	
	 iptopo = 0;
	 break;
      } /*switch*/
   }

   else if (pb->event->xbutton.button==Button2){
      /* Middle Button -> toggle slider/selector window on/off ***/
      if (pb->state == LUI_BUTTON_OUT)
         pb->state = LUI_BUTTON_IN;
      else
         pb->state = LUI_BUTTON_OUT;
      LUI_ColorHilite = Color[row][col];

      switch (col) {
      case ISOSURF:
	 if (ip3d>=0)
	    XUnmapWindow(wdpy,IsoWindow);
	 if (ip3d == row)
	   ip3d = -1;
	 else {
	    ip3d = row;
	    map_isosurf_window( ip3d );
	 }
	 break;
      case HSLICE:
	 if (ip2d>=0)
	    XUnmapWindow(wdpy,HSliceWindow);
	 if (ip2d==row)
            ip2d = -1;
	 else {
	    ip2d = row;
	    map_hslice_window( ip2d );
         }
	 break;
      case VSLICE:
	 if (ipxs>=0)
	    XUnmapWindow(wdpy,VSliceWindow);
	 if (ipxs==row)
            ipxs = -1;
	 else {
	    ipxs = row;
	    map_vslice_window( ipxs );
         }
	 break;
      case CHSLICE:
	 if (iphs==row) {
	    XUnmapWindow(wdpy,RGBWindow);
	    iphs = -1;
	 }
	 else {
	    iphs = row;
	    map_chslice_window( iphs );
         }
	 ipvs = -1;
	 iptopo = 0;
	 break;
      case CVSLICE:
	 if (ipvs==row) {
	    XUnmapWindow(wdpy,RGBWindow);
	    ipvs = -1;
	 }
	 else {
	    ipvs = row;
	    map_cvslice_window( ipvs );
         }
	 iphs = -1;
	 iptopo = 0;
	 break;

      }
   }
   else if (pb->event->xbutton.button==Button3) {
      /* Right Button -> color selector */
      if (pb->state == LUI_BUTTON_OUT)
         pb->state = LUI_BUTTON_IN;
      else
         pb->state = LUI_BUTTON_OUT;
      LUI_ColorHilite = Color[row][col];

      Last_button = pb;
      map_color_window( &Color[row][col] );

   }

   Redraw = 1;
}



static map_hwind_window( ws )
int ws;
{
   iphw = ws;

   XUnmapWindow( wdpy, WindWindow );
   sprintf( WindLabStr[0], "H-Wind %d scale: ", ws+1 );
   LUI_LabelChangeText( "WindLabel", 1, WindLabStr );
   sprintf( WindLab2Str[0], "Density:" );
   LUI_LabelChangeText( "WindLabel2", 1, WindLab2Str );
   sprintf( WindScaleStr[0], "%f", HWindScale[ws] );
   LUI_LabelChangeText( "WindScaleLabel", 1, WindScaleStr );
   sprintf( WindDensityStr[0], "%f", HWindDensity[ws] );
   LUI_LabelChangeText( "WindDensityLabel", 1, WindDensityStr );
   XMapWindow( wdpy, WindWindow );
}


static map_vwind_window( ws )
int ws;
{
   ipvw = ws;

   XUnmapWindow( wdpy, WindWindow );
   sprintf( WindLabStr[0], "V-Wind %d scale: ", ws+1 );
   LUI_LabelChangeText( "WindLabel", 1, WindLabStr );
   sprintf( WindLab2Str[0], "Density:" );
   LUI_LabelChangeText( "WindLabel2", 1, WindLab2Str );
   sprintf( WindScaleStr[0], "%f", VWindScale[ws] );
   LUI_LabelChangeText( "WindScaleLabel", 1, WindScaleStr );
   sprintf( WindDensityStr[0], "%f", VWindDensity[ws] );
   LUI_LabelChangeText( "WindDensityLabel", 1, WindDensityStr );
   XMapWindow( wdpy, WindWindow );
}


static map_isosurf_window( i )
int i;
{
   LUI_SliderChange( IsoSlider, ParmName[i],
		    MinVal[i], MaxVal[i], IsoLevel[i] );
   XMapWindow(wdpy,IsoWindow);
}



static map_hslice_window( i )
int i;
{
   sprintf( HSliceLabStr[0], "%s Horizontal Slice Interval: ", ParmName[i] );
   LUI_LabelChangeText( "HSliceLabel", 1, HSliceLabStr );
/*   sprintf( HSliceIntStr[0], "%10.5f", HSliceInterval[i] );*/
   sprintf( HSliceIntStr[0], "%f", HSliceInterval[i] );
   LUI_LabelChangeText( "HSliceIntLabel", 1, HSliceIntStr );
   XMapWindow(wdpy,HSliceWindow);
}



static map_vslice_window( i )
int i;
{
   sprintf( VSliceLabStr[0], "%s Vertical Slice Interval: ", ParmName[i] );
   LUI_LabelChangeText( "VSliceLabel", 1, VSliceLabStr );
/*   sprintf( VSliceIntStr[0], "%10.5f", VSliceInterval[i] );*/
   sprintf( VSliceIntStr[0], "%f", VSliceInterval[i] );
   LUI_LabelChangeText( "VSliceIntLabel", 1, VSliceIntStr );
   XMapWindow(wdpy,VSliceWindow);
}


static map_chslice_window( i )
int i;
{
   char str[100];

   XUnmapWindow( wdpy, RGBWindow );
   sprintf(str, "Horiz. %s ", ParmName[i] );
   map_cw( RGBWindow, MinVal[i], str, MaxVal[i] );
}


static map_cvslice_window( i )
int i;
{
   char str[100];

   XUnmapWindow( wdpy, RGBWindow );
   sprintf(str, "Vert. %s ", ParmName[i] );
   map_cw( RGBWindow, MinVal[i], str, MaxVal[i] );
}



static map_color_window( colorptr )
unsigned long *colorptr;
{
   unsigned long color;

   if (colorptr==ColorPtr) {
      /* unmap */
      ColorPtr = NULL;
      XUnmapWindow(wdpy,ColorWindow);
   }
   else {
      /* map */
      XUnmapWindow(wdpy,ColorWindow);
      XSync( wdpy, 0 );

      ColorPtr = colorptr;
      color = *colorptr;
      red_slider->value   = (color >> 16 & 0xff) / 255.0;
      green_slider->value = (color >> 8  & 0xff) / 255.0;
      blue_slider->value  = (color       & 0xff) / 255.0;
      trans_slider->value = (color >> 24 & 0xff) / 255.0;

      XMapWindow(wdpy,ColorWindow);
   }
}



static map_topo_color_window()
{
   XUnmapWindow(wdpy,RGBWindow);
   map_cw( RGBWindow, MinTopoHgt, "Topography Height", MaxTopoHgt );
}



/****************************************************************/
/*** CALL-BACK FUNCTIONS FOR ISOSURFACE, SLICE WINDOW WIDGETS ***/
/****************************************************************/


/*** isosurface_cb ****************************************************
   This is the call-back function for the isosurface level widget.
**********************************************************************/
static isosurface_cb(s)
LUI_SLIDER *s;
{
   if (IsoLevel[ip3d]!=s->value) {
      IsoLevel[ip3d] = s->value;
   }
}


static newsurf_cb()
{
   if (ip3d>-1) {
      request_isosurface( CurTime, ip3d );
   }
}



/*** hslice_cb ********************************************************
   This is the call-back function for the horizontal contour slice
   interval widget.
**********************************************************************/
static int hslice_cb( lab )
LUI_LABEL *lab;
{
   int hslice_cb2();

   if (lab->event->type==KeyPress) {
      LUI_LabelUserInput( lab, 0, hslice_cb2, lab->event, 1 );
      XSync( wdpy, 0 );
   }
   else if (lab->event->type==EnterNotify) {
      /* clear the type-in */
      strcpy( HSliceIntStr[0], "" );
      LUI_LabelChangeText( "HSliceIntLabel", 1, HSliceIntStr );
      LUI_LabelHighlight( "HSliceIntLabel", 1 );
      XSync( wdpy, 0 );
   }
   else if (lab->event->type==LeaveNotify) {
      /* restore the type-in */
/*      sprintf( HSliceIntStr[0], "%10.5f", HSliceInterval[ip2d] );*/
      sprintf( HSliceIntStr[0], "%f", HSliceInterval[ip2d] );
      LUI_LabelChangeText( "HSliceIntLabel", 1, HSliceIntStr );
      LUI_LabelHighlight( "HSliceIntLabel", 0 );
      XSync( wdpy, 0 );
   }

}


/*** hslice_cb2 *******************************************************
   This function is called when the user has entered a new value into
   the 2-D contour interval widget.
**********************************************************************/
static int hslice_cb2( valstr )
char *valstr;
{
   float f;

   if (ip2d>=0) {
      sscanf( valstr, "%f", &f );  /* convert from ascii to float */
      if (HSliceInterval[ip2d]!=f) {
	 HSliceInterval[ip2d] = f;
	 request_hslice( CurTime, ip2d );
      }
   }
}



/*** vslice_cb ********************************************************
   This is the call-back function for the vertical contour slice
   interval widget.
**********************************************************************/
static int vslice_cb( lab )
LUI_LABEL *lab;
{
   int vslice_cb2();

   if (lab->event->type==KeyPress) {
      LUI_LabelUserInput( lab, 0, vslice_cb2, lab->event, 1 );
   }
   else if (lab->event->type==EnterNotify) {
      /* clear the type-in */
      strcpy( VSliceIntStr[0], "" );
      LUI_LabelChangeText( "VSliceIntLabel", 1, VSliceIntStr );
      LUI_LabelHighlight( "VSliceIntLabel", 1 );
   }
   else if (lab->event->type==LeaveNotify) {
      /* restore the type-in */
/*      sprintf( VSliceIntStr[0], "%10.5f", VSliceInterval[ipxs] );*/
      sprintf( VSliceIntStr[0], "%f", VSliceInterval[ipxs] );
      LUI_LabelChangeText( "VSliceIntLabel", 1, VSliceIntStr );
      LUI_LabelHighlight( "VSliceIntLabel", 0 );
   }

}



/*** vslice_cb2 *******************************************************
   This function is called when the user has entered a new value into
   the X-Section contour interval widget.
**********************************************************************/
static int vslice_cb2( valstr )
char *valstr;
{
   float f;
   if (ipxs>=0) {
      sscanf( valstr, "%f", &f );  /* convert from ascii to float */
      if (VSliceInterval[ipxs]!=f) {
	 VSliceInterval[ipxs] = f;
	 request_vslice( CurTime, ipxs );
      }
   }
}



/*** windscale_cb *****************************************************
   This is the call-back function for the wind vector scale widget.
**********************************************************************/
static int windscale_cb( lab )
LUI_LABEL *lab;
{
   int windscale_cb2();

   if (lab->event->type==KeyPress) {
      LUI_LabelUserInput( lab, 0, windscale_cb2, lab->event, 1 );
   }
   else if (lab->event->type==EnterNotify) {
      /* clear the type-in */
      strcpy( WindScaleStr[0], "" );
      LUI_LabelChangeText( "WindScaleLabel", 1, WindScaleStr );
      LUI_LabelHighlight( "WindScaleLabel", 1 );
   }
   else if (lab->event->type==LeaveNotify) {
      /* restore the type-in */
      if (iphw>=0)
	 sprintf( WindScaleStr[0], "%f", HWindScale[iphw] );
      else
	 sprintf( WindScaleStr[0], "%f", VWindScale[ipvw] );
      LUI_LabelChangeText( "WindScaleLabel", 1, WindScaleStr );
      LUI_LabelHighlight( "WindScaleLabel", 0 );
   }

}



/*** windscale_cb2 ****************************************************
   This function is called when the user has entered a new value into
   the wind scale widget.
**********************************************************************/
static int windscale_cb2( valstr )
char *valstr;
{
   float f;

   if (iphw>=0) {
      sscanf( valstr, "%f", &f );  /* convert from ascii to float */
      if (HWindScale[iphw]!=f) {
	 HWindScale[iphw] = f;
	 request_hwindslice( CurTime, iphw );
      }
   }
   else if (ipvw>=0) {
      sscanf( valstr, "%f", &f );  /* convert from ascii to float */
      if (VWindScale[ipvw]!=f) {
	 VWindScale[ipvw] = f;
	 request_vwindslice( CurTime, ipvw );
      }
   }
}


/*** winddensity_cb *****************************************************
   This is the call-back function for the wind vector density widget.
**********************************************************************/
static int winddensity_cb( lab )
LUI_LABEL *lab;
{
   int winddensity_cb2();

   if (lab->event->type==KeyPress) {
      LUI_LabelUserInput( lab, 0, winddensity_cb2, lab->event, 1 );
   }
   else if (lab->event->type==EnterNotify) {
      /* clear the type-in */
      strcpy( WindDensityStr[0], "" );
      LUI_LabelChangeText( "WindDensityLabel", 1, WindDensityStr );
      LUI_LabelHighlight( "WindDensityLabel", 1 );
   }
   else if (lab->event->type==LeaveNotify) {
      /* restore the type-in */
      if (iphw>=0)
	 sprintf( WindDensityStr[0], "%f", HWindDensity[iphw] );
      else
	 sprintf( WindDensityStr[0], "%f", VWindDensity[ipvw] );
      LUI_LabelChangeText( "WindDensityLabel", 1, WindDensityStr );
      LUI_LabelHighlight( "WindDensityLabel", 0 );
   }

}



/*** winddensity_cb2 ****************************************************
   This function is called when the user has entered a new value into
   the wind density widget.
**********************************************************************/
static int winddensity_cb2( valstr )
char *valstr;
{
   float d;

   sscanf( valstr, "%f", &d );  /* convert from ascii to float */
   if (d<0.0 || d>1.0)
     d = 1.0;

   if (iphw>=0) {
      if (HWindDensity[iphw]!=d) {
	 HWindDensity[iphw] = d;
	 request_hwindslice( CurTime, iphw );
      }
   }
   else if (ipvw>=0) {
      if (VWindDensity[ipvw]!=d) {
	 VWindDensity[ipvw] = d;
	 request_vwindslice( CurTime, ipvw );
      }
   }
}



/********************************************/
/*** CALL-BACK FUNCTIONS FOR RGBA SLIDERS ***/
/********************************************/

static red_cb(s)
LUI_SLIDER *s;
{
   int red;

   if (ColorPtr) {
      red = 255 * (s->value);
      *ColorPtr = (*ColorPtr & 0xff00ffff) | (red << 16);
      /* update the button highlight color */
      LUI_ColorHilite = *ColorPtr;
      /* refreshes the button color */
      if (Last_button)
 	 LUI_ButtonDraw(Last_button);

      Redraw = 1;
   }
}


static green_cb(s)
LUI_SLIDER *s;
{
   int green;

   if (ColorPtr) {
      green = 255 * (s->value);
      *ColorPtr = (*ColorPtr & 0xffff00ff) | (green << 8);
      LUI_ColorHilite = *ColorPtr;
      if (Last_button)
	 LUI_ButtonDraw(Last_button);

      Redraw = 1;
   }
}


static blue_cb(s)
LUI_SLIDER *s;
{
   int blue;

   if (ColorPtr) {
      blue = 255 * (s->value);
      *ColorPtr = (*ColorPtr & 0xffffff00) | blue;
      LUI_ColorHilite = *ColorPtr;
      if (Last_button)
	 LUI_ButtonDraw(Last_button);

      Redraw = 1;
   }
}


static trans_cb(s)
LUI_SLIDER *s;
{
   int trans, i;

   if (ColorPtr) {
      /* change transparency of a color */
      trans = 255 * (s->value);
      *ColorPtr = (*ColorPtr & 0x00ffffff) | (trans << 24);

      /* check if changing color slice transparency */
      for (i=0;i<NumParms;i++) {
	 if (ColorPtr == &Color[i][CHSLICE]) {
	    cw_set_trans( CHSliceColors[i], trans );
	    break;
	 }
	 if (ColorPtr == &Color[i][CVSLICE]) {
	    cw_set_trans( CVSliceColors[i], trans );
	    break;
	 }
      }
      Redraw = 1;
   }
}




/************************************/
/***  Variable Cloning Functions  ***/
/************************************/

static int clone_window_mapped = 0;



/* called when the [CLONE] button's pressed */
static clone_cb()
{
   clone_window_mapped = !clone_window_mapped;

   if (clone_window_mapped) {
      /* check for room in tables for another variable */
      if (NumParms+NumClones==MAXPARMS) {
	 alert("Cannot make anymore clones.","");
	 return;
      }

      XMapWindow( wdpy, CloneWindow );
   }
   else {
      XUnmapWindow( wdpy, CloneWindow );
   }      
}



/* called when a button in the clone window is selected */
static do_clone_cb( b )
LUI_BUTTON *b;
{
   if (b->index>-1) {
      make_clone( b->index );
   }

   XUnmapWindow( wdpy, CloneWindow );
   clone_window_mapped = 0;
   Redraw = 1;
}



/*** make_clone *******************************************************
   This function actually clones the variable and adds a new row of
   buttons to the control panel.
   Input:  var_to_clone - which variable in [0..NumParms-1] to clone.
   Return:  number of the new varaiable or -1 if we hit the maximum
            allowable number of variables.
**********************************************************************/
int make_clone( var_to_clone )
int var_to_clone;
{
   int i, j, x, y, w, oldh, newh;
   char label[100];

   if (NumParms+NumClones==MAXPARMS)
      return -1;

   LUI_ButtonPadQuery("VarButtons", &x, &y, &w, &oldh);
   LUI_ButtonPadReopen("VarButtons");
   i = NumParms+NumClones;

   /* make parmname and add ''s */
   strcpy( ParmName[i], ParmName[var_to_clone] );
   for (j=0;j<NumParms+NumClones;j++) {
      if (CloneTable[j]==var_to_clone)
	strcat( ParmName[i], "'" );
   }

   for (j=0;j<5;j++) {
      sprintf( label, "%s        %d", ParmName[i], j );
      LUI_ColorHilite = Color[i][j];
      Button[NumButtons++] = LUI_ButtonCreate( label, LUI_BUTTON_TOGGLE,
					      j, i, 72, i*5+j, button_cb );
   }
   LUI_ButtonPadClose("VarButtons");
   LUI_ButtonPadQuery("VarButtons", &x, &y, &w, &newh);
   CpHeight += (newh-oldh);
   LUI_ResizeWindow( CpWindow, CP_WIDTH, CpHeight );
   XSync( wdpy, 0 );

   CloneTable[i] = var_to_clone;
   MinVal[i] = MinVal[var_to_clone];
   MaxVal[i] = MaxVal[var_to_clone];

   IsoLevel[i] = MinVal[i];
   HSliceLevel[i] = (float) (Nl-1) / 2;
   HSliceInterval[i] = round( (MaxVal[i] - MinVal[i]) / 10.0 );
   VSliceR1[i] = (float) (Nr-1) / 2;
   VSliceC1[i] = 0.0;
   VSliceR2[i] = (float) (Nr-1) / 2;
   VSliceC2[i] = (float) (Nc-1);
   VSliceInterval[i] = round( (MaxVal[i] - MinVal[i]) / 10.0 );
   CHSliceLevel[i] = (float) (Nl-1) / 2;
   CVSliceR1[i] = (float) (Nr-1) / 2;
   CVSliceC1[i] = 0.0;
   CVSliceR2[i] = (float) (Nr-1) / 2;
   CVSliceC2[i] = (float) (Nc-1);

   NumClones++;
   return NumParms+NumClones-1;
}



/* toggle pretty rendering on/off */
static pretty_cb()
{
   PrettyFlag = !PrettyFlag;
   set_pretty( PrettyFlag );
   Redraw = 2;
}


/* turn pretty rendering off */
static pretty_off()
{
   if (PrettyFlag) {
      PrettyFlag = 0;
      set_pretty(0);
      pretty_button->state = LUI_BUTTON_OUT;
      LUI_ButtonDraw( pretty_button );
   }
}


/* toggle reverse video option */
static reverse_cb()
{
   unsigned long t = BgColor;
   BgColor = BoxColor;
   BoxColor = LabelColor = t;
   Redraw = 1;
}


/* toggle information display */
static info_cb()
{
   DisplayInfo = !DisplayInfo;
   Redraw = 1;
}



/*** These functions are called when the F1, F2, F3, etc. function ***/
/*** keys are called.  You can put whatever you want in them to    ***/
/*** make macros, shortcuts, etc.                                  ***/

func1()
{
   /* raise control panel window */
   XRaiseWindow( wdpy, CpWindow );
}

func2()
{
}

func3()
{
}

func4()
{
}

func5()
{
}

func6()
{
}

func7()
{
}

func8()
{
}

func9()
{
}



/*** get_user_input ***************************************************
   Monitor the mouse for clicks, drags, etc.
   Input:  X events.
   Output:  none.
   Side-effects:  The CTM and other global viewing parameters will
                  be updated.
**********************************************************************/
void get_user_input()
{
   XEvent  pe, ne;
   int  current_x, current_y;
   static int  start_x, start_y, p1 = 0, p2 = 0, p3 = 0;
   float  anglex, angley, scale, transx, transy;
   static float  zmin = 1.0, zdelt;  /* for clipping */
   static Matrix  ctm;
   static float zoom;
   Matrix  mat;
   int ev, i;

#ifdef SPACEBALL
   prompt_spaceball();
#endif

   /* check for event, return if none in queue */
   if (XPending(wdpy)) {
      XNextEvent(wdpy, &pe);
   }
   else if (XPending(gdpy)) {
      XNextEvent(gdpy, &pe);
   }
   else {
#ifdef SPACEBALL
      if (check_spaceball())
	Redraw =1;
#endif
      return;
   }

   if (pe.xany.window==mainwin) {

      if (MouseMode==MM_SLICE && slice_event(pe)) {
	 pretty_off();
	 Redraw = 1;
      }
      else if (MouseMode==MM_LABEL && labeling_event(pe) ) {
	 pretty_off();
	 Redraw = 1;
      }
      else if (MouseMode==MM_TRAJ && (ev=cursor_event(pe))) {
	 if (ev==2) {
	    if (GoTime) {
	       for (i=0;i<NumTimes;i++) {
		  request_traj( CursorR, CursorC, CursorL, i,
			       iptraj, RibbonFlag, TrajStep, TrajLength );
	       }
	    }
	    else {
	       /* one traj */
	       request_traj( CursorR, CursorC, CursorL, CurTime,
			    iptraj, RibbonFlag, TrajStep, TrajLength );
	    }
	 }
	 pretty_off();
	 Redraw = 1;
      }
      else if (pe.type == UnmapNotify) {
	 /* Go to sleep if iconified */
	 do {
            if (XPending(gdpy))
	       XNextEvent(gdpy, &pe);
	    else
	       XNextEvent(wdpy, &pe);
	 } while (pe.type!=MapNotify);
      }
      
      else if (pe.type==Expose || pe.type==ConfigureNotify
	       || pe.type==VisibilityNotify) {
	 Redraw = 3;
      }
      
      /* Button press */
      else if (pe.type == ButtonPress) {
	 start_x = pe.xbutton.x;
	 start_y = pe.xbutton.y;
	 mat_copy(ctm, CTM);
	 if (pe.xbutton.button==Button1 && !p2 && !p3)
	   p1 = 1;
	 else if (pe.xbutton.button==Button2 && !p1 && !p3) {
	    p2 = 1;
	    zdelt = 0.0;
	 }
	 else if (pe.xbutton.button==Button3 && !p1 && !p2)
	   p3 = 1;
	 pretty_off();
      } /* if pe.type==ButtonPress */
      
      /* Button down and pointer moving */
      else if (pe.type == MotionNotify) {
	 current_x = pe.xmotion.x;
	 current_y = pe.xmotion.y;
	 
	 if (p1) {
	    /* rotate */
	    angley = (start_x-current_x) * 200.0 / (float) Width;
	    anglex = (start_y-current_y) * 200.0 / (float) Height;
	    make_matrix( anglex,angley,0.0, 1.0, 0.0,0.0,0.0, mat );
	    mat_mul( CTM, ctm, mat );
	    mat_inv( CTMinv, CTM );
	    pretty_off();
	    Redraw = 1;
	 }
	 else if (p2) {
	    /* clip */
	    zdelt = (start_x - current_x) / (float) Width;
	    set_frontclip( zmin + zdelt );
	    /* zoom */
	    /*
	      scale = 1.0 + ((start_y-current_y)/500.0);
	      */
	    scale = exp( (start_y-current_y) / (float) Height );
	    make_matrix( 0.0,0.0,0.0, scale, 0.0,0.0,0.0, mat );
	    mat_mul( CTM, ctm, mat );
	    mat_inv( CTMinv, CTM );
	    pretty_off();
	    Redraw = 1;
	 }
	 else if (p3) {
	    /* translate */
	    transx = (start_x - current_x) * -2.0 / (float) Width;
	    transy = (start_y - current_y) * 2.0 / (float) Height;
	    make_matrix( 0.0,0.0,0.0, 1.0, transx,transy,0.0, mat );
	    mat_mul( CTM, ctm, mat );
	    mat_inv( CTMinv, CTM );
	    pretty_off();
	    Redraw = 1;
	 }
      } /* if pe.type==MotionNotify */
      
      /* Button release */
      else if (pe.type == ButtonRelease) {
	 if (pe.xbutton.button == Button1)
	   p1 = 0;
	 else if (pe.xbutton.button == Button2) {
	    p2 = 0;
	    zmin = zmin + zdelt;
	 }
	 else if (pe.xbutton.button == Button3)
	   p3 = 0;
	 Redraw = 1;
      } /* if pe.type==ButtonRelease */
      
      /* Key press */
      else if (pe.type == KeyPress) {
	 KeySym keysym;
	 XComposeStatus compose;
	 char buff[20];
	 XLookupString(&pe, buff, 4, &keysym, &compose);
	 if (keysym == XK_F1) func1();
	 else if (keysym == XK_F2) func2();
	 else if (keysym == XK_F3) func3();
	 else if (keysym == XK_F4) func4();
	 else if (keysym == XK_F5) func5();
	 else if (keysym == XK_F6) func6();
	 else if (keysym == XK_F7) func7();
	 else if (keysym == XK_F8) func8();
	 else if (keysym == XK_F9) func9();
      }

      /* Flush remaining motion events in event queue */
      while (QLength(gdpy) > 0) {
	 XPeekEvent(gdpy, &ne);
	 
	 /* Don't flush if its a button press or release */
	 if (ne.type == ButtonPress || ne.type == ButtonRelease ||
	     ne.xany.window != mainwin)
	   break;
	 
	 /* Get the next event */
	 XNextEvent(gdpy, &ne);
      }
      
   } /* endif pe.xany.window==mainwin */
   
   else if (pe.xany.window==RGBWindow) {
      int trans, r;
      r = 0;
      if (iptopo) {
	 /* changing topo colors */
	 r = cw_event( pe, topo_color_table, 0xff );
	 if (r==1) {
	    recolor_topo( topo_color_table, 256 );
	 }
	 else if (r==2) {
	    init_topo_color_table( topo_color_table, 256 );
	    recolor_topo( topo_color_table, 256 );
	    redraw_cw( RGBWindow, topo_color_table );
	 }
      }
      else if (iphs>=0) {
	 /* changing horizontal slice colors */
	 trans = (Color[iphs][CHSLICE] & 0xff000000) >> 24;
	 r = cw_event( pe, CHSliceColors[iphs], trans );
	 if (r==2) {
	    init_ctable( CHSliceColors[iphs], trans );
	    redraw_cw( RGBWindow, CHSliceColors[iphs] );
	 }
      }
      else if (ipvs>=0) {
	 /* changing verrtical slice colors */
	 trans = (Color[ipvs][CVSLICE] & 0xff000000) >> 24;
	 r = cw_event( pe, CVSliceColors[ipvs], trans );
	 if (r==2) {
	    init_ctable( CVSliceColors[ipvs], trans );
	    redraw_cw( RGBWindow, CVSliceColors[ipvs] );
	 }
      }
      if (r) {
	 pretty_off();
	 Redraw = 1;
      }

   }

   else {
      /* the input event is not in mainwin, pass it on to LUI handler */
      if (pe.type==ButtonPress)
	pretty_off();
      LUI_EventDispatch(&pe);
   }

}

